/// Functions for performing cryptographic operations on the main data structures.

use super::*;
use openssl::{
    error::ErrorStack,
    encrypt::{Encrypter, Decrypter},
    pkey::{PKey, Public, Private},
    symm::{Cipher, encrypt as openssl_encrypt, decrypt as openssl_decrypt},
};
use serde_block_tree::{to_vec, from_vec};
use serde::de::{DeserializeOwned};

/// Data that may or may not be encrypted.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Cryptotext<T: Serialize> {
    /// The inner value of `T` is in plaintext.
    Plain(T),
    /// The inner value of `T` is in ciphertext.
    Cipher(Vec<u8>),
}

pub trait Hydrate<const N: usize> {
    fn hydrate(template: Self, data: [u8; N]) -> Self;
}

/// Errors that can occur during cryptographic operations.
#[derive(Debug)]
pub enum Error {
    NoReadCap,
    NoKeyAvailable,
    MissingPrivateKey,
    Encryption(String),
    Decryption(String),
}

pub type Result<T> = std::result::Result<T, Error>;

/// A cryptographic hash.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
pub enum Hash {
    Sha2_256([u8; 32]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; 64]),
}

/// A cryptographic signature.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Signature {
    #[serde(with = "BigArray")]
    Ed25519([u8; 64]),
}

/// A cryptographic key.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Key {
    // For keys that contain IVs, the key always comes first. The function `key_len` can be used
    // to get the length of the key part of a variant's array.

    /// A key for the AES 256 cipher in Cipher Block Chaining mode. Note that this includes the
    /// initialization vector, so that a value of this variant contains all the information needed
    /// to fully initialize a cipher context.
    Aes256Cbc {
        key: [u8; 32],
        iv: [u8; 16],
    },
    /// A key for the AES 256 cipher in counter mode.
    Aes256Ctr([u8; 32]),
    Ed25519 {
        public: [u8; 32],
        private: Option<[u8; 32]>
    },
}

enum EncryptionAlgo<'a> {
    Symmetric {
        cipher: Cipher,
        key: &'a [u8],
        iv: Option<&'a [u8]>
    },
    Asymmetric(PKey<Public>),
}

enum DecryptionAlgo<'a> {
    Symmetric {
        cipher: Cipher,
        key: &'a [u8],
        iv: Option<&'a [u8]>
    },
    Asymmetric(PKey<Private>),
}

impl From<ErrorStack> for Error {
    fn from(error: ErrorStack) -> Error {
        Error::Encryption(error.to_string())
    }
}

impl Key {
    /// Returns the key part of this `Key`. Note that if this `Key` is for an asymmetric algorithm,
    /// then this returns the *private* key (if it is present).
    fn key_slice(&self) -> Option<&[u8]> {
        match self {
            Key::Aes256Cbc { key, .. } => Some(key),
            Key::Aes256Ctr(key) => Some(key),
            Key::Ed25519 { private, .. } => match private {
                Some(key) => Some(key.as_slice()),
                None => None,
            },
        }
    }

    /// Returns the IV part of this`Key`.
    fn iv_slice(&self) -> Option<&[u8]> {
        match self {
            Key::Aes256Cbc { iv, .. } => Some(iv),
            _ => None,
        }
    }
}

impl<'a> TryFrom<&'a Key> for EncryptionAlgo<'a> {
    type Error = Error;
    fn try_from(key: &'a Key) -> Result<EncryptionAlgo<'a>> {
        match key {
            Key::Aes256Cbc { key: key_slice, iv } => Ok(EncryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_cbc(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Aes256Ctr(key_slice) => Ok(EncryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_ctr(),
                key: key_slice,
                iv: None
            }),
            Key::Ed25519 { public, .. } => {
                let pkey = PKey::public_key_from_der(public.as_slice())
                    .map_err(|err| Error::Encryption(err.to_string()));
                Ok(EncryptionAlgo::Asymmetric(pkey?))
            },
        }
    }
}

impl<'a> EncryptionAlgo<'a> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        match self {
            EncryptionAlgo::Symmetric { cipher, key, iv } => {
                openssl_encrypt(*cipher, key, *iv, slice).map_err(Error::from)
            },
            EncryptionAlgo::Asymmetric(pkey) => {
                let encrypter = Encrypter::new(pkey).map_err(Error::from)?;
                let buffer_len = encrypter.encrypt_len(slice).map_err(Error::from)?;
                let mut ciphertext = vec![0; buffer_len];
                let ciphertext_len = encrypter.encrypt(slice, &mut ciphertext)
                    .map_err(Error::from)?;
                ciphertext.truncate(ciphertext_len);
                Ok(ciphertext)
            }
        }
    }
}

impl<'a> TryFrom<&'a Key> for DecryptionAlgo<'a> {
    type Error = Error;
    fn try_from(key: &'a Key) -> Result<DecryptionAlgo<'a>> {
        match key {
            Key::Aes256Cbc { key: key_slice, iv } => Ok(DecryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_cbc(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Aes256Ctr(key_slice) => Ok(DecryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_ctr(),
                key: key_slice,
                iv: None
            }),
            Key::Ed25519 { private, .. } => {
                let private = private.ok_or(Error::MissingPrivateKey)?;
                let pkey = PKey::private_key_from_der(private.as_slice())
                    .map_err(|err| Error::Decryption(err.to_string()));
                Ok(DecryptionAlgo::Asymmetric(pkey?))
            }
        }
    }
}

impl<'a> DecryptionAlgo<'a> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        match self {
            DecryptionAlgo::Symmetric { cipher, key, iv } => {
                openssl_decrypt(*cipher, key, *iv, slice).map_err(Error::from)
            },
            DecryptionAlgo::Asymmetric(pkey) => {
                let decrypter = Decrypter::new(pkey).map_err(Error::from)?;
                let buffer_len = decrypter.decrypt_len(slice).map_err(Error::from)?;
                let mut plaintext = vec![0; buffer_len];
                let plaintext_len = decrypter.decrypt(slice, &mut plaintext)
                    .map_err(Error::from)?;
                plaintext.truncate(plaintext_len);
                Ok(plaintext)
            },
        }
    }
}

#[allow(dead_code)]
pub(crate) fn encrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Cipher(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Plain(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = encrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Cipher(new_body);
    block.read_caps.insert(principal_owned, encrypt(&block_key, key)?);
    Ok(VersionedBlock::V0(block))
}

#[allow(dead_code)]
pub(crate) fn decrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Plain(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Cipher(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = decrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Plain(new_body);
    block.read_caps.insert(principal_owned, Cryptotext::Plain(block_key));
    Ok(VersionedBlock::V0(block))
}

fn encrypt<T: Serialize>(value: &T, key: &Key) -> Result<Cryptotext<T>> {
    let data = to_vec(value).map_err(|err| Error::Encryption(err.to_string()))?;
    let vec = encrypt_slice(&data, key);
    Ok(Cryptotext::Cipher(vec?))
}

fn decrypt<T: Serialize + DeserializeOwned>(cryptotext: Cryptotext<T>, key: &Key) -> Result<T> {
    let data = match cryptotext {
        Cryptotext::Plain(value) => return Ok(value),
        Cryptotext::Cipher(data) => data
    };
    let vec = decrypt_slice(&data, key);
    from_vec(&vec?).map_err(|err| Error::Decryption(err.to_string()))
}

fn encrypt_slice(plaintext: &[u8], key: &Key) -> Result<Vec<u8>> {
   let algo = EncryptionAlgo::try_from(key)?;
   algo.encrypt(plaintext)
}

fn decrypt_slice(ciphertext: &[u8], key: &Key) -> Result<Vec<u8>> {
    let algo = DecryptionAlgo::try_from(key)?;
    algo.decrypt(ciphertext)
}

#[cfg(test)]
mod tests {
    use super::*;

    const KEY: [u8; 32] = [
        0xB2, 0xB3, 0xDA, 0x5A, 0x1A, 0xF6, 0xB3, 0x78, 0x30, 0xAB, 0x1D, 0x33, 0x33, 0xE7, 0xE3,
        0x5B, 0xBB, 0xF9, 0xFE, 0xD0, 0xC1, 0xF7, 0x90, 0x34, 0x69, 0xB7, 0xE7, 0xC6, 0x1C, 0x46,
        0x85, 0x48,
    ];

    const IV: [u8; 16] = [
        0x60, 0xE0, 0xBE, 0x45, 0xA9, 0xAB, 0x5D, 0x99, 0x3B, 0x3A, 0x1E, 0x54, 0x18, 0xE0, 0x46,
        0xDE,
    ];

    #[test]
    fn key_aes256cbc_key_slice_is_correct() -> Result<()> {
        let key = Key::Aes256Cbc { key: KEY, iv: IV };
        let key_part = key.key_slice().ok_or(Error::NoKeyAvailable);
        assert_eq!(KEY, key_part?);
        Ok(())
    }

    #[test]
    fn key_aes256cbc_iv_slice_is_correct() {
        let key = Key::Aes256Cbc { key: KEY, iv: IV };
        let iv_part = key.iv_slice().unwrap();
        assert_eq!(IV, iv_part);
    }

    /// Tests that validate the dependencies of this module.
    mod dependency_tests {
        /// This test validates that data decrypted with the `Crypter` API matches data that was
        /// previously encrypted using it.
        #[test]
        fn crypter() {
            use super::*;
            let expected = b"We attack at the crack of noon!";
            let cipher = Cipher::aes_256_cbc();

            let ciphertext = openssl_encrypt(cipher, &KEY, Some(&IV), expected).unwrap();
            let actual = openssl_decrypt(cipher, &KEY, Some(&IV), ciphertext.as_slice()).unwrap();

            assert_eq!(expected, actual.as_slice());
        }
    }
}
