/// Tests which ensure that the main data structures can be round-tripped faithfully.

use super::*;
use crypto::{Hash, Signature, Key};
use serde_block_tree::{Error, Result, from_vec, to_vec};

static PRINCIPAL: [u8; 32] = [
    0x75, 0x28, 0xA9, 0xE0, 0x9D, 0x24, 0xBA, 0xB3, 0x79, 0x56, 0x15, 0x68, 0xFD, 0xA4, 0xE2, 0xA4,
    0xCF, 0xB2, 0xC0, 0xE3, 0x96, 0xAE, 0xA2, 0x6E, 0x45, 0x15, 0x50, 0xED, 0xA6, 0xBE, 0x6D, 0xEC,
];

static PAYLOAD: [u8; 128] = [
    0x39, 0x79, 0x1A, 0x0D, 0x8E, 0x6C, 0xF5, 0x4B, 0xF3, 0xA4, 0x75, 0xC4, 0x44, 0x73, 0x58, 0x58,
    0x97, 0x14, 0x64, 0xE0, 0xC6, 0xFE, 0xCB, 0xCF, 0xBE, 0x67, 0x49, 0x49, 0x40, 0xAE, 0x71, 0x5A,
    0x94, 0x7E, 0x6C, 0x4B, 0xDE, 0x33, 0x22, 0x75, 0xD8, 0x54, 0x23, 0x37, 0xFD, 0x1A, 0x68, 0x4A,
    0x5F, 0xB5, 0xB3, 0xC9, 0x9A, 0x72, 0x7C, 0xF4, 0x3C, 0xAB, 0xED, 0x97, 0x87, 0x63, 0xBB, 0xD9,
    0x8B, 0x11, 0xD3, 0xC1, 0x4C, 0x9A, 0x09, 0x0E, 0x7C, 0x10, 0x65, 0x9B, 0x8F, 0x35, 0xEB, 0x51,
    0x19, 0xD7, 0x6E, 0xA3, 0xC9, 0x64, 0xE2, 0x54, 0x84, 0x5F, 0xA1, 0x8B, 0x63, 0x0C, 0xC3, 0x9D,
    0xBE, 0xBB, 0x7F, 0x31, 0x1D, 0x59, 0xE2, 0x68, 0xC4, 0x5B, 0x37, 0x77, 0x04, 0xAD, 0x44, 0x75,
    0xEE, 0x1F, 0x84, 0x17, 0xA2, 0x74, 0xC3, 0xC3, 0xD5, 0x2F, 0x70, 0x74, 0xFE, 0xD8, 0x2C, 0x29,
];

static SIGNATURE: [u8; 64] = [
    0x2E, 0x19, 0x8E, 0xCC, 0x2D, 0xE1, 0x0E, 0x42, 0x3F, 0xBB, 0x89, 0x3D, 0x07, 0xAB, 0x57, 0xBF,
    0xEB, 0xB1, 0xA7, 0x23, 0xF9, 0xD2, 0xC3, 0x3F, 0x7A, 0x3C, 0xD0, 0x31, 0x38, 0x01, 0x33, 0x1F,
    0x07, 0x8D, 0x68, 0x0C, 0x6B, 0xF1, 0xBA, 0xD3, 0xD4, 0xAE, 0xAD, 0x1A, 0xF7, 0x5D, 0x2C, 0xEF,
    0x1F, 0x5C, 0x50, 0xE9, 0xFA, 0x8A, 0xDB, 0xB2, 0x4C, 0xC6, 0x9B, 0x06, 0x5F, 0xFB, 0xE3, 0xDA,
];

static KEY: [u8; 32] = [
    0xB2, 0xB3, 0xDA, 0x5A, 0x1A, 0xF6, 0xB3, 0x78, 0x30, 0xAB, 0x1D, 0x33, 0x33, 0xE7, 0xE3, 0x5B,
    0xBB, 0xF9, 0xFE, 0xD0, 0xC1, 0xF7, 0x90, 0x34, 0x69, 0xB7, 0xE7, 0xC6, 0x1C, 0x46, 0x85, 0x48,
];

static LONG_KEY: [u8; 64] = [
    0x65, 0x4E, 0x78, 0x5C, 0x8E, 0xEE, 0x9E, 0x57, 0x58, 0x8D, 0xDE, 0x9E, 0x6D, 0x84, 0xF4, 0x68,
    0x7F, 0x9E, 0x48, 0xC4, 0x6F, 0x1F, 0x4F, 0x83, 0x46, 0x68, 0xB2, 0xC1, 0x46, 0xD0, 0x06, 0xF2,
    0x23, 0xB9, 0xD4, 0x66, 0x4B, 0xED, 0x65, 0x92, 0xF1, 0xA2, 0x04, 0x1A, 0xD9, 0xF6, 0xA2, 0xAE,
    0x87, 0x3E, 0xBF, 0x10, 0x4A, 0x29, 0x28, 0x83, 0x2A, 0x8F, 0xC5, 0x30, 0x86, 0x11, 0x48, 0x52,
];

#[test]
fn roundtrip_fragment_record() -> Result<()> {
    let expected = FragmentRecord::new(229, Hash::Sha2_256(PRINCIPAL));
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}

#[test]
fn roundtrip_directory() -> Result<()> {
    let principal = Principal(Hash::Sha2_256(PRINCIPAL));
    let mut child = HashMap::new();
    child.insert(FragmentSerial(0), FragmentRecord::new(0, principal.0));
    let mut children = HashMap::new();
    children.insert(".metadata".to_string(), child);
    let expected = Directory {
        nodes: vec![Principal(Hash::Sha2_256(PRINCIPAL))],
        children
    };
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}

#[test]
fn roundtrip_fragment() -> Result<()> {
    let body = Vec::from(PAYLOAD);
    let expected = Fragment::new("apps/bodhi", 42, body)
        .map_err(|err| Error::Message(err.to_string()))?;
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}

fn make_write_cap() -> Result<WriteCap> {
    Ok(WriteCap {
        issued_to: Principal(Hash::Sha2_256(PRINCIPAL)),
        issued_by: Principal(Hash::Sha2_256(PRINCIPAL)),
        path: Path::try_from("contacts/emergency").map_err(|err| Error::Message(err.to_string()))?,
        expires: Epoch(1649904316),
        signing_key: Key::Ed25519 { public: KEY, private: None },
        signature: Signature::Ed25519(SIGNATURE),
        next: Some(Box::from(WriteCap {
            issued_to: Principal(Hash::Sha2_256(PRINCIPAL)),
            issued_by: Principal(Hash::Sha2_256(PRINCIPAL)),
            path: Path::try_from("contacts").map_err(|err| Error::Message(err.to_string()))?,
            expires: Epoch(1649994316),
            signing_key: Key::Ed25519 { public: KEY, private: None },
            signature: Signature::Ed25519(SIGNATURE),
            next: None,
        }))
    })
}

#[test]
fn roundtrip_write_cap() -> Result<()> {
    let expected = make_write_cap()?;
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}

fn make_read_cap() -> ReadCap {
    ReadCap::new(Hash::Sha2_256(PRINCIPAL), Cryptotext::Plain(Key::Aes256Ctr(KEY)))
}

#[test]
fn roundtrip_read_cap() -> Result<()> {
    let expected = make_read_cap();
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}

fn make_block() -> Result<Block> {
    let mut read_caps = HashMap::new();
    {
        let read_cap = make_read_cap();
        read_caps.insert(read_cap.issued_to, read_cap.key);
    }
    let write_cap = make_write_cap()?;
    Ok(Block {
            path: Path::try_from("apps/verse").map_err(|err| Error::Message(err.to_string()))?,
            read_caps,
            write_cap,
            body: Cryptotext::Plain(Vec::from(PAYLOAD)),
            signature: Signature::Ed25519(SIGNATURE)
    })
}

#[test]
fn roundtrip_versioned_block() -> Result<()> {
    let expected = VersionedBlock::V0(make_block()?);
    let ser_result = to_vec(&expected);
    let de_result = from_vec(&ser_result?);
    assert_eq!(expected, de_result?);
    Ok(())
}